package jwiley.chp4;

import java.lang.*;
import java.lang.Math;
import java.util.Random;
import java.util.Date;

/**
 * A simple implementation of the dining philosophers problem which
 * results in deadlock.
 *
 * The table class represents the table that the philosophers will sit at.
 *
 * @author  A. Saganich
 * @version 1.0
 *
 */
class Table 
{
	final int FREE = -1;
	int philosopherCount = 0;
	int forks[];
	Philosopher phils[];
	/** The names of a few of my favorite philosophers */
	String names[] = {"Plato",
					  "Aristotle",
					  "Socrates",
					  "Alan Turing",
					  "Mr Bill",
					  "Sluggo",
					  "Mr Ed",
					  "Decartes",
					  "St. Augustine",
					  "Occam",
					  "Nietzsche"};
					  
	/* Create a table and seat philosphers at it */					  
	Table (int howMany)
	{
		philosopherCount = howMany;
		forks = new int[howMany];
		phils = new Philosopher[howMany];
		for ( int i = 0; i < howMany; i++)
		{	
			forks[i] = FREE;
			phils[i] = new Philosopher();
			phils[i].init(this,names[i],i);
			phils[i].start();
		}
	}
	
	/* Put down the allocated forks */	
	synchronized public void  putDownForks(int which)
	{
		int left = which;
		int right = (which + 1) % philosopherCount;
		forks[left] = FREE;
		forks[right] = FREE;
	}
	/* Pick up the right fork */	
	synchronized public boolean  pickUpRightFork(int which)
	{
		int right = (which + 1) % philosopherCount;
		if (forks[right] != FREE)
			return false;
		
		forks[right] = which;
		return true;
	}
	/* Pick up the right left */	
	synchronized public boolean  pickUpLeftFork(int which)
	{
		int left = which;
		if (forks[left] !=  FREE)
			return false;
		
		forks[left] = which;
		return true;
	}
	
	/** Start the simulation and then let it run for a while */
	public static void main (String[] args)
	{
		Table table = new Table(5); // Five philosophers
		 
		try { Thread.sleep(10000*5); } catch (Exception e) { System.exit(1); };
		
		for ( int i = 0; i < 5; i++)
		{
			table.phils[i].stop();
		}		
	}
	
} // end class Table

/**
 * Each philosopher is a thread.
 * 
 * @author  A. Saganich
 * @version 1.0
 *
 */
class Philosopher extends Thread
{
	final	int timeSlice = 10000;  // Max time anything can happen for.
	int		thinkTime;
	int 	eatTime;
	int 	talkTime;
	String 	name;
	int 	which;
	Random	dice;
	Table	table = null;
	
	public Philosopher()
	{	
		init(null,"Unknown",-1);
	}
	public Philosopher(Table table, String name, int which)
	{
		init(table,name,which);
	}
	
	int getRand()
	{
		return Math.abs(dice.nextInt()) %timeSlice;
	}
	int getRand(int modifier)
	{
		return Math.abs(dice.nextInt()) %modifier;
	}
	public void init(Table table, String name, int which)
	{
		dice = new Random();
		this.name = name;
		this.which = which;
		this.table = table;
		thinkTime = getRand();
		eatTime   = getRand();
		talkTime  = getRand();
	}
	
	void thinking()
	{
		Date currentTime = new Date(System.currentTimeMillis());
		int howLong = getRand(thinkTime);
		System.out.println(name + " is thinking for " + howLong + "ms at " + currentTime);
		try { sleep(howLong); } catch (Exception e) {} ;
	}
	
	void eating()
	{
		Date currentTime = new Date(System.currentTimeMillis());
		int howLong = getRand(eatTime);
		System.out.println(name + " is eating for " + howLong + "ms at " + currentTime);
		try { sleep(howLong); } catch (Exception e) {} ;
	}
	
	void talking()
	{
		Date currentTime = new Date(System.currentTimeMillis());
		int howLong = getRand(talkTime);
		System.out.println(name + " is talking for " + howLong + "ms at " + currentTime);
		try { sleep(howLong); } catch (Exception e) {} ;
	}
	
	/** 
	 * The run method does all the work by thinking for a while
	 * and then talking for a while to build up an appetite and
	 * then obtaining forks and finally eating.
	 * The process then repeats itself.
	 */

	public void run()
	{
		Date currentTime = new Date(System.currentTimeMillis());
		System.out.println(name +" sat in seat " + which + " at " + currentTime);
		do
		{
		
			System.out.println(name + " attempting to pick up left fork");
			while (table.pickUpLeftFork(which) == false) 
			{
				try { sleep(100); } catch (Exception e) {} ;
			}
			System.out.println(name + " obtained left fork");
			System.out.println(name + " attempting to pick up right fork");
			while ( table.pickUpRightFork(which) == false)
			{
				try { sleep(100); } catch (Exception e) {} ;
			}
			System.out.println(name + " obtained right fork");
			eating();
			table.putDownForks(which);
			thinking();
			talking();
		}while (true);
	}
} // End class Philosopher